/***************************************************************************
 *   Copyright (C) 2006 by André Lindhjem <belgarat@sdf.lonestar.org>,     *
 *                         Kjetil Holien <kjetil.holien@gmail.com>,        *
 *                         Terje Risa <terje.risa@gmail.com> &             *
 *                         Øyvind Nerbråten <oyvind@nerbraten.com>         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************
 *
 * \file control.c
 * \author André Lindhjem, Kjetil Holien, Terje Risa & Øyvind Nerbråten
 * \date 23.02.2006
 * 
 * Datastructure and methods for controlling it.
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>
#include <pthread.h>
#include <libxml2/libxml/xmlreader.h>

#include "control.h"
#include "nccparser.h"
#include "ncxparser.h"
#include "nccsmilparser.h"
#include "ncxsmilparser.h"
#include "report.h"
#include "common.h"
#include "audio.h"
#include "libdaisy.h"
#include "snprintf/snprintf.h"


/* Test the xml reader is present */
#ifdef LIBXML_READER_ENABLED



/* ************************************************** *
 * 			Static function declarations 			  *
 * ************************************************** */

/**
 * Initilizes a new authorNode by setting all its pointers to NULL and all its values to 0.
 * @param authorNode - a pointer to the authorNode to initalize.
 */
static void initAuthorNode (struct AuthorNode* authorNode);

/**
 * Initilizes a new smilNode by setting all its pointers to NULL and all its values to 0.
 * @param smilNode - a pointer to the smilNode to initalize.
 */
static void initSmilNode (struct SmilNode* smilNode);

/**
 * Initilizes a new  passage node by setting all its pointers to NULL and all its values to 0.
 * @param node - a pointer to the node to initalize.
 */
static void initNode (struct Node* node);

/**
 * Removes all smilnodes in the linked list in the main datastructure
 * of the loaded daisy DTB and deallocates their memory.
 * @param daisydata - a pointer to the datastructure initialized by parseInit ().
 */
static void flushSmilNodes (struct DaisyData *daisydata);

/**
 * Removes all passage nodes in the linked list in the main datastructure
 * of the loaded daisy DTB and deallocates their memory.
 * @param daisydata - a pointer to the datastructure initialized by parseInit ().
 */
static void flushNodes (struct DaisyData *daisydata);

/**
 * Removes the bookinfo structure containing the metadata and the linked list containing
 * all the authors in the main datastructure of the loaded daisy DTB and deallocates their memory.
 * @param daisydata - a pointer to the datastructure initialized by parseInit ().
 */
static void flushBookInfo (struct DaisyData *daisydata);



/* ************************************************** *
 * 			Static function definitions 			  *
 * ************************************************** */

/**
 * Initilizes a new authorNode by setting all its pointers to NULL and all its values to 0.
 * @param authorNode - a pointer to the authorNode to initalize.
 */
static void initAuthorNode (struct AuthorNode* authorNode)
{
	authorNode->name = NULL;
	authorNode->audiofilename = NULL;
	authorNode->audioStartPos = NULL;
	authorNode->audioStopPos = NULL;
	authorNode->image = NULL;
	authorNode->next = NULL;
	authorNode->prev = NULL;
}

/**
 * Initilizes a new smilNode by setting all its pointers to NULL and all its values to 0.
 * @param smilNode - a pointer to the smilNode to initalize.
 */
static void initSmilNode (struct SmilNode* smilNode)
{
	smilNode->id = NULL;
	smilNode->anchor = NULL;
	smilNode->fragmentIdentifier = NULL;
	smilNode->textpassage = NULL;
	smilNode->audiofilename = NULL;
	smilNode->audioStartPos = NULL;
	smilNode->audioStopPos = NULL;
	smilNode->image = NULL;
	smilNode->header = NULL;
	smilNode->next = NULL;
	smilNode->prev = NULL;
	smilNode->optional = 0;
}

/**
 * Initilizes a new passage node by setting all its pointers to NULL and all its values to 0.
 * @param node - a pointer to the node to initalize.
 */
static void initNode (struct Node* node)
{
	node->audiofilename = NULL;
	node->audioStartPos = NULL;
	node->audioStopPos = NULL;
	node->fragmentIdentifier = NULL;
	node->image = NULL;
	node->textPassage = NULL;
	node->textfilename = NULL;
	node->next = NULL;
	node->prev = NULL;
}

/**
 * Removes all smilnodes in the linked list in the main datastructure
 * of the loaded daisy DTB and deallocates their memory.
 * @param daisydata - a pointer to the datastructure initialized by parseInit ().
 */
static void flushSmilNodes (struct DaisyData *daisydata)
{
	struct SmilNode *tmp = daisydata->smilHead->next;
	while (tmp != daisydata->smilTail)
	{
		if (tmp->id != NULL) 
			free (tmp->id);
		if (tmp->anchor != NULL)
			free (tmp->anchor);
		if (tmp->fragmentIdentifier != NULL)
			free (tmp->fragmentIdentifier);
		if (tmp->header != NULL)
			free (tmp->header);
		tmp = tmp->next;
		free (tmp->prev);
	}
	daisydata->smilHead->next = daisydata->smilTail;
	daisydata->smilTail->prev = daisydata->smilHead;
	daisydata->smilPos = daisydata->smilHead;
}

/**
 * Removes all passage nodes in the linked list in the main datastructure
 * of the loaded daisy DTB and deallocates their memory.
 * @param daisydata - a pointer to the datastructure initialized by parseInit ().
 */
static void flushNodes (struct DaisyData *daisydata)
{
	struct Node *tmp = daisydata->nodeHead->next;
	while (tmp != daisydata->nodeTail)
	{
		if (tmp->audiofilename != NULL)
			free (tmp->audiofilename);
		if (tmp->audioStartPos != NULL)
			free (tmp->audioStartPos);
		if (tmp->audioStopPos != NULL)
			free (tmp->audioStopPos);
		if (tmp->textPassage != NULL)
			free (tmp->textPassage);
		tmp = tmp->next;
		free (tmp->prev);	
	}
	daisydata->nodeHead->next = daisydata->nodeTail;
	daisydata->nodeTail->prev = daisydata->nodeHead;
	daisydata->nodePos = daisydata->nodeHead;
}

/**
 * Removes the bookinfo structure containing the metadata and the linked list containing
 * all the authors in the main datastructure of the loaded daisy DTB and deallocates their memory.
 * @param daisydata - a pointer to the datastructure initialized by parseInit ().
 */
static void flushBookInfo (struct DaisyData *daisydata)
{
	struct AuthorNode *tmp;
	
	if (daisydata->bookInfo->titleText != NULL)
	{ 
		free (daisydata->bookInfo->titleText);
		daisydata->bookInfo->titleText = NULL;
	}
	if (daisydata->bookInfo->titleAudiofilename != NULL)
	{
		free (daisydata->bookInfo->titleAudiofilename);
		daisydata->bookInfo->titleAudiofilename = NULL;
	}
	if (daisydata->bookInfo->titleAudioStartPos != NULL)
	{
		free (daisydata->bookInfo->titleAudioStartPos);
		daisydata->bookInfo->titleAudioStartPos = NULL;
	}
	if (daisydata->bookInfo->titleAudioStopPos != NULL)
	{
		free (daisydata->bookInfo->titleAudioStopPos);
		daisydata->bookInfo->titleAudioStopPos = NULL;
	}
	if (daisydata->bookInfo->titleImage != NULL)
	{
		free (daisydata->bookInfo->titleImage);
		daisydata->bookInfo->titleImage = NULL;
	}
	if (daisydata->bookInfo->totalTime != NULL)
	{
		free (daisydata->bookInfo->totalTime);
		daisydata->bookInfo->totalTime = NULL;
	}			
			
	tmp = daisydata->authorHead->next;
	while (tmp != daisydata->authorTail)
	{
		if (tmp->name != NULL) 
			free (tmp->name);
		if (tmp->audiofilename != NULL)
			free (tmp->audiofilename);
		if (tmp->audioStartPos != NULL)
			free (tmp->audioStartPos);
		if (tmp->audioStopPos != NULL)
			free (tmp->audioStopPos);
		if (tmp->image != NULL)
			free (tmp->image);
		tmp = tmp->next;
		free (tmp->prev);
	}
	daisydata->authorHead->next = daisydata->authorTail;
	daisydata->authorHead->prev = daisydata->authorHead;
	daisydata->bookInfo->author = daisydata->authorHead;
}



/* ************************************************** *
 * 			Global function definitions 			  *
 * ************************************************** */

/**
 * Initializes the main datastructure of the loaded daisy DTB.
 * @returns a pointer to the main datastructure for loading a daisy DTB.
 */
struct DaisyData *parseInit (void)
{	
	/* The main datastructure of a loaded daisy DTB. */
	struct DaisyData *daisydata = (struct DaisyData *) malloc (sizeof(struct DaisyData));
	
	
	/*
     * This initialize the library and check potential ABI mismatches
     * between the version it was compiled for and the actual shared
     * library used.
     */
	LIBXML_TEST_VERSION
	
		
	/* Initalizing the datastucture for storing book meta data */
	daisydata->bookInfo = (struct BookInfo *) malloc (sizeof (struct BookInfo));
	daisydata->bookInfo->titleText = NULL;
	daisydata->bookInfo->titleAudiofilename = NULL;
	daisydata->bookInfo->titleAudioStartPos = NULL;
	daisydata->bookInfo->titleAudioStopPos = NULL;
	daisydata->bookInfo->titleImage = NULL;
	daisydata->bookInfo->totalTime = NULL;
	daisydata->authorHead = (struct AuthorNode *) malloc (sizeof (struct AuthorNode));
	initAuthorNode (daisydata->authorHead);	
	daisydata->authorTail = (struct AuthorNode *) malloc (sizeof (struct AuthorNode));
	initAuthorNode (daisydata->authorTail);
	daisydata->authorHead->next = daisydata->authorTail;
	daisydata->authorHead->prev = daisydata->authorHead;
	daisydata->bookInfo->author = daisydata->authorHead;

	/* Initializing the head and the tail of the smil linked lists */
	daisydata->smilHead = (struct SmilNode *) malloc (sizeof (struct SmilNode));
	initSmilNode (daisydata->smilHead);
	daisydata->smilTail = (struct SmilNode *) malloc (sizeof (struct SmilNode));
	initSmilNode (daisydata->smilTail);
	daisydata->smilHead->next = daisydata->smilTail;
	daisydata->smilTail->prev = daisydata->smilHead;
	daisydata->smilPos = daisydata->smilHead;
	
	/* Initializing the head and the tail of the node linked lists */
	daisydata->nodeHead = (struct Node *) malloc (sizeof (struct Node));
	initNode (daisydata->nodeHead);
	daisydata->nodeTail = (struct Node *) malloc (sizeof (struct Node));
	initNode (daisydata->nodeTail);
	daisydata->nodeHead->next = daisydata->nodeTail;
	daisydata->nodeTail->prev = daisydata->nodeHead;
	daisydata->nodePos = daisydata->nodeHead;
	
	return daisydata;
}

/**
 * Frees all memory allocated by the parser and the main datastructure.
 * Terminates all parsing functionality.
 * @param daisydata - a pointer to the datastructure initialized by parseInit ().
 */
void parseTerminate (struct DaisyData *daisydata)
{
	
	/* Frees the nodelist including head and tail */
	flushNodes(daisydata);
	free (daisydata->nodeHead);
	free (daisydata->nodeTail);
	
	/* Frees the smillist including head and tail */
	flushSmilNodes(daisydata);
	free (daisydata->smilHead);
	free (daisydata->smilTail);
	
	/* Frees the book info including head and tail */
	flushBookInfo(daisydata);
	free (daisydata->authorHead);
	free (daisydata->authorTail);
	free (daisydata->bookInfo);
	
	/* Frees the main data struct */
	free (daisydata);
	
	/* Cleanup function for the XML library. */
	xmlCleanupParser ();
    
	/* This is to debug memory for regression tests */
	xmlMemoryDump ();
		
}

/**
 * Removes whitespaces, newlines and tabs before and after a sentence.
 * @param string - a pointer to the string to check.
 * @returns a pointer to the new string if the sentence contains any characters, 
 * otherwise NULL.
 */
char* removewhitespaces(char* string)
{	
	unsigned int start = 0; /* Counter variables	*/
	int end = 0;
	int i = 0;
	int n = 0;
	char* newstring = NULL; /* The new string */
	
	/* Checks the characters before the string */
	for (start = 0; start < strlen(string); start++)
	{	
		if (string[start] != ' ' && string[start] != '\t' && string[start] != '\n')
			break;
	}
	
	/* If the sentences only contains whitespaces,tabs or newlines, return NULL */
	if (start == strlen(string)) return NULL;
	
	/* Checks the characters from the end and backwards */
	for (end = strlen(string)-1; end >= 0; end--)
	{	
		if (string[end] != ' ' && string[end] != '\t' && string[end] != '\n')
			break;
	}
	
	/* Creates the new string with the non-characters stripped */	
	newstring = malloc(end-start+2);
	for (i = start; i <= end; i++)
	{
		newstring[n++] = string[i];
	}
	newstring[n] = '\0';	
	return newstring;
}

/**
 * Converts the string to lower case.
 * @param string - a pointer to the string which is going to be in lowercase.
 * @returns a pointer to the new string with lowercase, otherwise NULL.
 */
char* tolowercase(const char* string)
{
	unsigned int n = 0; /* Counter variable	*/
	char* newstring = NULL; /* The new string */
	
	newstring = malloc(strlen(string)+1);
	/* Goes through the string and lowercases each char */
	for (n = 0; n < strlen(string); n++)
	{
		newstring[n] = tolower(string[n]);
			
	}
	newstring[n] = '\0';
	return newstring;
}


/**
 * Converts min:sec.frac to sec.frac.
 * @param time - a pointer to the string to convert.
 * @returns a pointer to the string contianing the offset.
 */
char* getTime(char* time)
{
	char* tmp, *min, *sec, *frac = NULL;
	char* value = (char *) malloc(6);	
	int minutes = 0, seconds = 0;
	
	/* copy time to a tmp variable */
	tmp = (char *) malloc (strlen (time)+1);
	strcpy (tmp, time);
	/* tokenize in order to get minutes */
	min = strtok (tmp, ":");
	minutes = atoi (min);
	/* tokenize in order to get the chars after min: */
	tmp = strtok (NULL, ":");
	/* tokenize in order to get seconds */
	sec = strtok (tmp, ".");
	/* convert char to int in order to add minutes to seconds */
	seconds = atoi(sec);
	seconds = (minutes * 60) + seconds;
	/* tokenize in order to get fraction */
	frac = strtok (NULL, ".");	
	/* convert the int seconds to char */
	snprintf(value, sizeof (value), "%d", seconds);
	/* add .fractions */
	strcat (value, ".");
	strcat (value, frac);
	return value;
}

/**
 * Adds a new author to the main data structures author list.
 * @param daisydata - a pointer to the datastructure initialized by parseInit ().
 */
void addNewAuthor (struct DaisyData *daisydata)
{	
	/* allocating memory for the new smilnode */
	struct AuthorNode *tmpAuthor = (struct AuthorNode *) malloc (sizeof(struct AuthorNode));
	
	/* initializes the authorNode */
	initAuthorNode(tmpAuthor);
	
	/* making a new author node after the head - already other authors in the list */
	if (daisydata->authorHead->next != daisydata->authorTail)
	{ 
		tmpAuthor->next  = daisydata->authorHead->next;
		tmpAuthor->prev  = daisydata->authorHead;
		daisydata->authorHead->next->prev  = tmpAuthor;
		daisydata->authorHead->next  = tmpAuthor;
	}
	/* no other authors in the list */
	else
	{		
		daisydata->authorHead->next = tmpAuthor;
		daisydata->authorTail->prev = tmpAuthor;
		tmpAuthor->next = daisydata->authorTail;
		tmpAuthor->prev = daisydata->authorHead;
	}
	daisydata->bookInfo->author = tmpAuthor;
}
	
/**
 * Adds a new SmilNode at the end of the linked smil list in the main datastructure
 * of the loaded daisy DTB.
 * @param daisydata - a pointer to the datastructure initialized by parseInit ().
 * @returns a pointer to the new smilNode.
 */
struct SmilNode* addNewSmilNode (struct DaisyData *daisydata)
{
	struct SmilNode *tmp;
	
	/* allocating memory for the new smilnode */
	struct SmilNode *smilNode = (struct SmilNode *) malloc (sizeof(struct SmilNode));
	
	/* initializes the smilNode */
	initSmilNode (smilNode);
	
	/* adding the empty smilnode til the end of the linked list */
	tmp = daisydata->smilTail->prev;
	tmp->next = smilNode;
	smilNode->prev = tmp;
	smilNode->next = daisydata->smilTail;
	daisydata->smilTail->prev = smilNode;
	
	return smilNode;
}

/**
 * Adds a new passage node at the end of the linked node list in the main datastructure
 * of the loaded daisy DTB.
 * @param daisydata - a pointer to the datastructure initialized by parseInit ().
 * @returns a pointer to the new node.
 */
struct Node* addNewNode (struct DaisyData *daisydata)
{
	struct Node *tmp;
	
	/* allocating memory for the new smilnode */
	struct Node *node = (struct Node *) malloc (sizeof(struct Node));
	
	/* initializes the node */
	initNode(node);
	
	/* adding the empty smilnode til the end of the linked list */
	tmp = daisydata->nodeTail->prev;
	tmp->next = node;
	node->prev = tmp;
	node->next = daisydata->nodeTail;
	daisydata->nodeTail->prev = node;
	
	return node;
}

/**
 * Parses a NCC or a NCX file (*.ncc | ncx.*) and fills the smilNode list with
 * the smil file information representing the spine of the daisy DTB.
 * @param daisydata - a pointer to the datastructure initialized by parseInit ().
 * @param filename - a pointer to a string containing the path and filename to the ncc/ncx file
 * to parse.
 * @returns 1 in case of success and -1 in case of error
 */
int parse (struct DaisyData *daisydata, char* filename)
{
	char buffer[STRLEN];
	int i = 0, j = 0, n = 0, ret = 0;
	i = (int) strlen(filename);
	
	/* make sure the datastructure is empty */
	flushNodes (daisydata);
	flushSmilNodes (daisydata);
	flushBookInfo (daisydata);
	
	/* checks wheter the file_name ends with .ncx or .NCX */
	if((filename[i-1] == 'x' && filename[i-2] == 'c' && filename[i-3] == 'n' && filename[i-4] == '.') 
	|| (filename[i-1] == 'X' && filename[i-2] == 'C' && filename[i-3] == 'N' && filename[i-4] == '.'))
	{	
		/* parse NCX, must extract the path */
		daisydata->format = 1;
		/* find the path */
		for(i = (int) strlen(filename)-1; i > 0; i--)
		{
			if(filename[i] == '/')
				break;
		}
		/* stores the path */
		for (j = 0; j <= i; j++)
		{
			buffer[n++] = filename[j];
		}
		buffer[n] = '\0';
		daisydata->path = (char*) malloc (sizeof (buffer)+1);
		assert (daisydata->path != NULL);	/** Make sure path isn't NULL */
		strcpy (daisydata->path, buffer);
		ret = parseNCX (daisydata, filename);	
	}
	/* 2.02 Daisy book must be ../ncc.html or ../NCC.HTML */
	else if((filename[i-1] == 'l' && filename[i-2] == 'm' && filename[i-3] == 't' && filename[i-4] == 'h'
	&&	filename[i-5] == '.' && filename[i-6] == 'c' && filename[i-7] == 'c' && filename[i-8] == 'n' && filename[i-9] == '/') 
	|| (filename[i-1] == 'L' && filename[i-2] == 'M' && filename[i-3] == 'T' && filename[i-4] == 'H'
	&&	filename[i-5] == '.' && filename[i-6] == 'C' && filename[i-7] == 'C' && filename[i-8] == 'N' && filename[i-9] == '/'))
	{
		/* we got a 2.02 Daisy book -> parseNCC */
		daisydata->format = 0;
		
		/* stores the path */
		for (j = 0; j <= (int) strlen(filename)-9; j++)
		{
			buffer[n++] = filename[j];
		}
		buffer[n] = '\0';
		daisydata->path = (char*) malloc (sizeof (buffer)+1);
		assert (daisydata->path != NULL);	/** make sure path isn't NULL */
		strcpy (daisydata->path, buffer);
		ret = parseNCC (daisydata, filename);
	}
	else
	{
		/*TODO: Error callback */
		return -1;
	}
	return ret;
}

/**
 * Function for seeking in the data structure. It can seek paragraphs (nodes) and
 * chapters (smilnode).
 * @param daisydata - a pointer to the datastructure initialized by parseInit ().
 * @param type - an daisy_seek_option, the type of seek you want to performe.
 * @returns 0 if end_of_book, 1 if success and -1 on error.
 */
int seek (struct DaisyData *daisydata, int type)
{
	int ret = 1;
	
	if (daisydata->smilHead->next == daisydata->smilTail) return 0;
	
	switch (type)
	{
		case DAISY_SEEK_PREV_PASSAGE:
			if (daisydata->nodePos == daisydata->nodeHead || daisydata->nodePos == daisydata->nodeHead->next)
			{
				if (daisydata->smilPos == daisydata->smilHead)
				{
					daisydata->smilPos = daisydata->smilHead->next;
					flushNodes (daisydata);
					if(daisydata->format == 0)
						ret = parseNCCSmil(daisydata);	
					else
						ret = parseNCXSmil(daisydata);
					if (ret != 1) return ret;
					daisydata->nodePos = daisydata->nodeHead->next;
					return ret;	
				}
				else if (daisydata->smilPos == daisydata->smilHead->next)	
				{
					daisydata->nodePos = daisydata->nodeHead->next;
					return ret;	
				}
				else
				{
					daisydata->smilPos = daisydata->smilPos->prev;
					flushNodes(daisydata);
					if(daisydata->format == 0)
						ret = parseNCCSmil(daisydata);	
					else
						ret = parseNCXSmil(daisydata);
					if (ret != 1) return ret;
					daisydata->nodePos = daisydata->nodeTail->prev;
					return ret;	
				}
			}
			else
			{
				daisydata->nodePos = daisydata->nodePos->prev;
				return ret;
			}
		
		case DAISY_SEEK_PREV_CHAPTER:
			if (daisydata->smilPos == daisydata->smilHead)
			{
				daisydata->smilPos = daisydata->smilHead->next;
				flushNodes(daisydata);
				if(daisydata->format == 0)
					ret = parseNCCSmil(daisydata);	
				else
					ret = parseNCXSmil(daisydata);
				if (ret != 1) return ret;
				daisydata->nodePos = daisydata->nodeHead->next;
				return ret;
			}
			else if (daisydata->smilPos == daisydata->smilHead->next)
			{
				daisydata->nodePos = daisydata->nodeHead->next;
				return ret;	
			}
			else
			{
				daisydata->smilPos = daisydata->smilPos->prev;
				flushNodes(daisydata);
				if(daisydata->format == 0)
					ret = parseNCCSmil(daisydata);	
				else
					ret = parseNCXSmil(daisydata);
				if (ret != 1) return ret;
				daisydata->nodePos = daisydata->nodeHead->next;
				return ret;	
			}
		
		case DAISY_SEEK_NEXT_CHAPTER:
			if (daisydata->smilPos == daisydata->smilTail)
			{
				return 0;
			}
			else if (daisydata->smilPos == daisydata->smilTail->prev)
			{
				daisydata->smilPos = daisydata->smilTail;
				return 0;
			}
			else {
				daisydata->smilPos = daisydata->smilPos->next;
				flushNodes(daisydata);
				if(daisydata->format == 0)
					ret = parseNCCSmil(daisydata);	
				else
					ret = parseNCXSmil(daisydata);
				if (ret != 1) return ret;
				daisydata->nodePos = daisydata->nodeHead->next;
				return ret;
			}
		
		case DAISY_SEEK_NEXT_PASSAGE:
			if (daisydata->nodePos == daisydata->nodeTail || daisydata->nodePos == daisydata->nodeTail->prev)
			{
				if (daisydata->smilPos == daisydata->smilTail)
				{
					daisydata->nodePos = daisydata->nodeTail;
					return 0;
				}
				else if (daisydata->smilPos == daisydata->smilTail->prev)
				{
					daisydata->smilPos = daisydata->smilTail;
					daisydata->nodePos = daisydata->nodeTail;
					return 0;	
				}
				else
				{
					daisydata->smilPos = daisydata->smilPos->next;
					flushNodes (daisydata);
					if(daisydata->format == 0)
						ret = parseNCCSmil(daisydata);	
					else
						ret = parseNCXSmil(daisydata);
					if (ret != 1) return ret;
					daisydata->nodePos = daisydata->nodeHead->next;
					return ret;
				}
			}
			else
			{
				daisydata->nodePos = daisydata->nodePos->next;
				return ret;	
			}
			
		case DAISY_SEEK_TO_BEGINNING:
			if (daisydata->smilPos != daisydata->smilHead->next)
			{
				daisydata->smilPos = daisydata->smilHead->next;
				flushNodes (daisydata);
				if(daisydata->format == 0)
					ret = parseNCCSmil(daisydata);	
				else
					ret = parseNCXSmil(daisydata);
				if (ret != 1) return ret;
			}
			daisydata->nodePos = daisydata->nodeHead->next;
			return ret;
		
		default:
			return -1;	
		
	}
}

/**
 * Function for getting the current playback position in the smil list.
 * @param daisydata - a pointer to the datastructure initialized by parseInit ().
 * @returns the current position or -1 on error.
 */
int getSmilPos (struct DaisyData *daisydata)
{
	int smilPos = 1;
	struct SmilNode *tmpSmil = daisydata->smilHead->next;
	
	if (daisydata->smilPos != daisydata->smilHead && daisydata->smilPos != daisydata->smilTail)
	{
		while (tmpSmil != daisydata->smilPos && tmpSmil != daisydata->smilTail)
		{
			smilPos++;
			tmpSmil = tmpSmil->next;
		}
		if (tmpSmil == daisydata->smilTail)
			return -1;
	} else
		return -1;
		
	return smilPos;	
}

/**
 * Function for getting the current playback position in the node list.
 * @param daisydata - a pointer to the datastructure initialized by parseInit ().
 * @returns the current position or -1 on error.
 */
int getNodePos (struct DaisyData *daisydata)
{
	int nodePos = 1;
	struct Node *tmpNode = daisydata->nodeHead->next;
	
	if (daisydata->nodePos != daisydata->nodeHead && daisydata->nodePos != daisydata->nodeTail)
	{
		while (tmpNode != daisydata->nodePos && tmpNode != daisydata->nodeTail)
		{
			nodePos++;
			tmpNode = tmpNode->next;
		}
		if (tmpNode == daisydata->nodeTail)
			return -1;
	} else
		return -1;
		
	return nodePos;
}

/**
 * Function to move to a spesific node in the smil list and parse it.
 * @param daisydata - a pointer to the datastructure initialized by parseInit ().
 * @param newSmilPos - the position in the linked smil list to move to.
 * @returns 1 on success and -1 on error.
 */
int gotoSmilPosition (struct DaisyData *daisydata, int newSmilPos)
{
	int i = 1;
	struct SmilNode *tmpSmil = daisydata->smilHead->next;
	
	/* check the parameter value newSmilPos */
	if (newSmilPos < 1)
		return -1;
	
	/* find the smilnode */
	for (i=1; i<newSmilPos && tmpSmil != daisydata->smilTail; i++)
	{
		tmpSmil = tmpSmil->next;
	}
	if (tmpSmil == daisydata->smilTail)
		return -1;
	
	/* set the new smil position */
	daisydata->smilPos = tmpSmil;
	
	/* parse the smilfile*/
	flushNodes (daisydata);
	if(daisydata->format == 0)
		i = parseNCCSmil(daisydata);	
	else
		i = parseNCXSmil(daisydata);
	if (i != 1) return i;
	
	return 1;
}

/**
 * Function to move to a spesific node in the smil list and parse it.
 * @param daisydata - a pointer to the datastructure initialized by parseInit ().
 * @param newNodePos - the position in the linked node list to move to.
 * @returns 1 on success and -1 on error.
 */
int gotoNodePosition (struct DaisyData *daisydata, int newNodePos)
{
	int i = 1;
	struct Node *tmpNode = daisydata->nodeHead->next;
	
	/* check the parameter value newNodePos */
	if (newNodePos < 1)
		return -1;
		
	/* find the node */
	for (i=1; i<newNodePos && tmpNode != daisydata->nodeTail; i++)
	{
		tmpNode = tmpNode->next;	
	}
	if (tmpNode == daisydata->nodeTail)
		return -1;
		
	/* set the new node position */
	daisydata->nodePos = tmpNode;
	
	return 1;
}

/**
 * Starts the audio thread. The audio playback is run as a separate thread with
 * callback functions.
 * @param daisy - a pointer to the daisy data stuct returned from daisy_init ().
 * @param done - a void pointer to a function which will be called when a audio passage 
 * is done playing with the daisy data struct as parameter.
 * @param error - a void pointer to a function which will be called when an error occurs 
 * in the audio thread with the audio data struct, the status of the audio thread and the error message
 * as parameters.
 * @param progress - a void pointer to a function which will be called continuously with the progress in 
 * ms as an parameter.
 * @return audio_data_t to the audio data struct or NULL on fail.
 */
audio_data_t start_audio (daisyplayer_t	daisy,
				  void			(*done)		(daisyplayer_t d),
				  void			(*error)	(void *data, enum daisy_status, const char *daisy_error_msg),
				  void			(*progress)	(void *data, long int progress_ms))
{
	audio_data_t ret;
	struct_daisyplayer_t *_daisy = daisy;
	
	ret = audio_initiate (_daisy, done, error, progress);

	if (DEBUG_DAISYPLAYER) report ("control.c: in start_audio", REP_DEBUG);

	if (pthread_create (&_daisy->daisy_book_data->thread_audio, NULL, &audio_thread, ret))
	{
		report ("Error creating audio thread", REP_CRITICAL);
		return NULL; /* fail */
	}
	else
	{
		if (DEBUG_DAISYPLAYER) report ("control.c: audio_thread started", REP_DEBUG);
	}

	return ret; /* succeed */
}

/**
 * Stops the audio thread.
 * @param daisy - a pointer the daisy data stuct returned from daisy_init ().
 * @param data - a pointer to the audio data struct.
 */
void stop_audio (daisyplayer_t daisy, audio_data_t data)
{
	struct_daisyplayer_t *_daisy = daisy;
	
	audio_terminate(data);		/* tell audiomodule to terminate and clean up	*/
	pthread_cancel (_daisy->daisy_book_data->thread_audio);
}

#endif
