/***************************************************************************
 *   Copyright (C) 2006 by André Lindhjem <belgarat@sdf.lonestar.org>,     *
 *                         Kjetil Holien <kjetil.holien@gmail.com>,        *
 *                         Terje Risa <terje.risa@gmail.com> &             *
 *                         Øyvind Nerbråten <oyvind@nerbraten.com>         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************
 *
 * \file control.h
 * \author André Lindhjem, Kjetil Holien, Terje Risa & Øyvind Nerbråten
 * \date 23.02.2006
 * 
 * Datastructure and methods for controlling it.
 */

 
#ifndef CONTROL_H_
#define CONTROL_H_

#include <pthread.h>
#include "libdaisy.h"
#include "audio.h"



/* ************************************************** *
 * 				Global data structures	 			  *
 * ************************************************** */

/**
 * The main datastructure of the loaded daisy DTB.
 */
struct DaisyData
{
	/* Pointers to head, tail and current position in the linked list containing all the smilfiles
	 * representing the spine of the loaded daisy DTB (this normally represents all chapters).
	 */
	struct SmilNode *smilHead, *smilTail, *smilPos;
		
	/* Pointers to head, tail and current position in the linked list containing all passage
	 * nodes of a parsed smilfile.
	 */
	struct Node *nodeHead, *nodeTail, *nodePos;
	
	/* Pointer to a struct containing the metadata of the loaded daisy DTB. */
	struct BookInfo *bookInfo;
	
	/* Pointer to head and tail of a linked list containing information off all authors
	 * of the loaded daisy DTB.
	 */
	struct AuthorNode *authorHead, *authorTail;
	
	/* Representing the book version format (e.g. 2.02) */
	int format;
	
	/* Used by the smilparsers to define if a smil file is already present */
	int smilAlreadyfound;
	
	/* The path to the directory where the daisy DTB is loaded. */
	char *path;
	
	/* Audio engine thread for the loaded daisy DTB. */
	pthread_t thread_audio;
};

/**
 * Datastructure for the meta data of the loaded daisy DTB. 
 */
struct BookInfo
{
	char *titleText; /* the title of the book */
	char *titleAudiofilename; /* the source to the audiofile containing the title */
	char *titleAudioStartPos; /* the time offset for where to start playback in the audio file */
	char *titleAudioStopPos; /* the time offset for where to stop playback in the audio file */
	char *titleImage; /* the source to the book image */
	char *totalTime; /* the total time of the book */
	struct AuthorNode *author; /* a linked list containing the authors with their respective text,audiofilename, audiostart/stopos and image */
};

/**
 * Datastructure for author information of the loaded daisy DTB.
 */
struct AuthorNode
{
	char *name; /* the title of the book */
	char *audiofilename; /* the source to the audiofile containing the title */
	char *audioStartPos; /* the time offset for where to start playback in the audio file */
	char *audioStopPos; /* the time offset for where to stop playback in the audio file */
	char *image; /* the source to the book image */
	struct AuthorNode *next; /* a pointer to the next authornode in the linked list */
	struct AuthorNode *prev; /* a pointer to the previous authornode in the linked list */
};

/**
 * Datastructure for keeping the information of the smilefiles, parsed from the ncc, using a linked list
 * <a> <div> <span>
 */
struct SmilNode
{
	int optional; /* marks optional passages */
	char *id; /* id for the SMIL file */ 
	char *anchor; /* the name of the SMIL file */
	char *header; /* the header of the contens of the smilfile */
	char *fragmentIdentifier; /* identifies a <par> or <text> element in the SMIL file */
	char *textpassage; /* a text corresponding to the smilNode */
	char *audiofilename; /* the filname of the corresponding audio file */
	char *audioStartPos; /* the time offset for where to start playback in the audio file */
	char *audioStopPos; /* the time offset for where to stop playback in the audio file */
	char *image; /* the source to the corresponding image */
	struct SmilNode *next; /* a pointer to the next smilnode in the linked list */
	struct SmilNode *prev;	/* a pointer to the previous smilnode in the linked list */
};

/**
 * The datastructure for a passage within the book. Holds the name and audio_play times of the corresponding
 * audio file and the text for the passage. Passages is stored in a linked list.
 */ 
struct Node
{ 
	/*int skippable;  true if the node is auto skippable (normally false) */
	char *audiofilename; /* the filname of the corresponding audio file */
	char *audioStartPos; /* the time offset for where to start playback in the audio file */
	char *audioStopPos; /* the time offset for where to stop playback in the audio file */
	char *textfilename; /* the file_name of the corresponding text file */
	char *fragmentIdentifier; /* identificator for the corresponding passage in the text file */
	char *textPassage; /* the corresponding text passage */
	char *image; /* the source to the corresponding image */
	struct Node *next; /* a pointer to the next node in the lined list */
	struct Node *prev; /* a pointer to the previous node int the linked list */
};



/* ************************************************** *
 * 			Global function declarations 			  *
 * ************************************************** */

/**
 * Initializes the main datastructure of the loaded daisy DTB.
 * @returns a pointer to the main datastructure for loading a daisy DTB.
 */
struct DaisyData *parseInit (void);

/**
 * Frees all memory allocated by the parser and the main datastructure.
 * Terminates all parsing functionality.
 * @param daisydata - a pointer to the datastructure initialized by parseInit ().
 */
void parseTerminate (struct DaisyData *daisydata);

/**
 * Removes whitespaces, newlines and tabs before and after a sentence.
 * @param string - a pointer to the string to check.
 * @returns a pointer to the new string if the sentence contains any characters, 
 * otherwise NULL.
 */
char* removewhitespaces(char* string);

/**
 * Converts the string to lower case.
 * @param string - a pointer to the string which is going to be in lowercase.
 * @returns a pointer to the new string with lowercase, otherwise NULL.
 */
char* tolowercase(const char* string);

/**
 * Converts min:sec.frac to sec.frac.
 * @param time - a pointer to the string to convert.
 * @returns a pointer to the string contianing the offset.
 */
char* getTime(char* time);

/**
 * Adds a new author to the main data structures author list.
 * @param daisydata - a pointer to the datastructure initialized by parseInit ().
 */
void addNewAuthor (struct DaisyData *daisydata);

/**
 * Adds a new SmilNode at the end of the linked smil list in the main datastructure
 * of the loaded daisy DTB.
 * @param daisydata - the datastructure initialized by parseInit ().
 * @returns a pointer to the new smilNode.
 */
struct SmilNode* addNewSmilNode (struct DaisyData *daisydata);

/**
 * Adds a new passage node at the end of the linked node list in the main datastructure
 * of the loaded daisy DTB.
 * @param daisydata - a pointer to the datastructure initialized by parseInit ().
 * @returns a pointer to the new node.
 */
struct Node* addNewNode (struct DaisyData *daisydata);

/**
 * Parses a NCC or a NCX file (*.ncc | ncx.*) and fills the smilNode list with
 * the smil file information representing the spine of the daisy DTB.
 * @param daisydata - a pointer to the datastructure initialized by parseInit ().
 * @param filename - a pointer to a string containing the path and filename to the ncc/ncx file
 * to parse.
 * @returns 1 in case of success and -1 in case of error
 */
int parse (struct DaisyData *daisydata, char* filename);

/**
 * Function for seeking in the data structure. It can seek paragraphs (nodes) and
 * chapters (smilnode).
 * @param daisydata - a pointer to the datastructure initialized by parseInit ().
 * @param type - an daisy_seek_option, the type of seek you want to performe.
 * @returns 0 if end_of_book, 1 if success and -1 on error.
 */
int seek (struct DaisyData *daisydata, int type);

/**
 * Function for getting the current playback position in the smil list.
 * @param daisydata - a pointer to the datastructure initialized by parseInit ().
 * @returns the current position or -1 on error.
 */
int getSmilPos (struct DaisyData *daisydata);

/**
 * Function for getting the current playback position in the node list.
 * @param daisydata - a pointer to the datastructure initialized by parseInit ().
 * @returns the current position or -1 on error.
 */
int getNodePos (struct DaisyData *daisydata);

/**
 * Function to move to a spesific node in the smil list and parse it.
 * @param daisydata - a pointer to the datastructure initialized by parseInit ().
 * @param newSmilPos - the position in the linked smil list to move to.
 * @returns 1 on success and -1 on error.
 */
int gotoSmilPosition (struct DaisyData *daisydata, int newSmilPos);

/**
 * Function to move to a spesific node in the smil list and parse it.
 * @param daisydata - a pointer to the datastructure initialized by parseInit ().
 * @param newNodePos - the position in the linked node list to move to.
 * @returns 1 on success and -1 on error.
 */
int gotoNodePosition (struct DaisyData *daisydata, int newNodePos);

/**
 * Starts the audio thread. The audio playback is run as a separate thread with
 * callback functions.
 * @param daisy - the daisy data stuct returned from daisy_init ().
 * @param done - a void pointer to a function which will be called when a audio passage 
 * is done playing with the daisy data struct as parameter.
 * @param error - a void pointer to a function which will be called when an error occurs 
 * in the audio thread with the audio data struct, the status of the audio thread and the error message
 * as parameters.
 * @param progress - a void pointer to a function which will be called continuously with the progress in 
 * ms as an parameter.
 * @return audio_data_t to the audio data struct or NULL on fail.
 */
audio_data_t start_audio (daisyplayer_t	daisy,
				  void (*done) (daisyplayer_t),
				  void (*error) (void *data, enum daisy_status, const char *),
				  void (*progress) (void *data, long int progress_ms));

/**
 * Stops the audio thread.
 * @param daisy - a pointer the daisy data stuct returned from daisy_init ().
 * @param data - a pointer to the audio data struct.
 */
void stop_audio (daisyplayer_t	daisy, audio_data_t data);


#endif /*CONTROL_H_*/
